/**
 * This JavaScript provides an abstracted method of dealing with "behind
 * the scenes" client-server requests (e.g. XMLHTTP/XMLHttpRequest/AJAX).
 *
 * $Horde: incubator/Horde_Form/js/ajax.js,v 1.4 2006/03/02 23:50:19 chuck Exp $
 *
 * Copyright 2005 Matt Warden <mwarden@gmail.com>
 *
 * See the enclosed file COPYING for license information (LGPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/lgpl.html.
 *
 * @author  Matt Warden <mwarden@gmail.com>
 * @package Horde_Form
 */

/**
 * This is the class that deals with remote requests.
 * This implementation uses XMLHttpRequest, but the
 * idea is that you could swap implementations without
 * changing any code that uses this class.
 *
 * @author Matt Warden <mwarden@gmail.com>
 * @access public
 */
function RemoteConnection(defaulturl, recurseOnChildren)
{
   // constants
   this.HTTP_STATUS_OK = 200;
   this.READYSTATE_COMPLETED = 4;
   this.NODE_TYPE_ELEMENT = 1;

   this.defaulturl = defaulturl;
   // array where we'll store our request objects
   this.aRequests = new Array();
   this.aRequests[0] = null;
   // our container to map elements to handling
   // functions
   this.hRespHandlers = new SimpleHashtable();
   // flag to track whether a wildcard handler
   // has been set.
   this.isWildcardSet = false;
   // whether or not to recurse on children of a
   // matched subtree that has already been passed
   // to an appropriate handler function.
   // the code is such because it is an optional
   // parameter and might not be defined (but if
   // it is defined and isn’t false then we also know
   // its value. So we just assign it literally.
   this.recurseOnChildren =
      recurseOnChildren ? true : false;
   // error container
   this.errors = new Array();


   this.getErrors = function ()
      {
          if (this.errors.length > 0) {
              return this.errors.join('\n');
          }
          else {
              return false;
          }
      }

   this.request = function(url, method, requestxml)
      {
         // set defaults omitted optional arguments
         if (!url) url = this.defaulturl;
         if (!method) method = "GET";
         if (!requestxml) requestxml = null;

// TODO: remove
document.getElementById('xmlinput').value = Sarissa.serialize(requestxml);

         this.defaulturl = url;

         // request object
         var req = null;

         // look for an empty spot in requests
         // array due to a deleted request. You
         // might consider moving this
         // functionality to library code and
         // wrapping the functionality in a
         // data structure.

         // default spot is at end
         var openIndex = this.aRequests.length;
         // look for closer spot
         for (var i=0; i<this.aRequests.length; i++)
         {
            if (this.aRequests[i] == null)
            {
               openIndex = i;
               break; // stop looping
            } // end if
         }

         // now make the request, if possible
         if (window.XMLHttpRequest)
         {
            // this might look odd, but it is
            // necessary because ‘this’ in event
            // handlers refers to the owner of the
            // fired event. See:
            // www.quirksmode.org/js/this.html
            var self = this;
            req = new XMLHttpRequest();
            req.onreadystatechange =
               function() {self.handle()};
            // add the element to the array before
            // doing anything that will fire
            // readyStateChange event. If we didn’t
            // do this now, we could be getting event
            // firings from request objects that we
            // can’t find in our requsts array, when we
            // go to handle the readyStateChange.
            this.aRequests[openIndex] = req;
            req.open(method, url, true);
            req.send(requestxml);
         }
         else if (window.ActiveXObject)
         {
            req = new ActiveXObject("Microsoft.XMLHTTP");
            if (req)
            {
               // this might look odd, but it is
               // necessary because ‘this’ in event
               // handlers refers to the owner of the
               // fired event. See:
               // www.quirksmode.org/js/this.html
               var self = this;
               req.onreadystatechange =
                  function() {self.handle()};
               // add the element to the array before
               // doing anything that will fire
               // readyStateChange event. If we didn’t
               // do this now, we could be getting event
               // firings from request objects that we
               // can’t find in our requsts array, when
               // we go to handle the readyStateChange.
               this.aRequests[openIndex] = req;
               req.open(method, url, true);
               req.send(requestxml);
            }
            else
            {
                return false; // indicate an error
            } // end if req created
         }
         else
         {
            // no support
            return false; // indicate an error
         } // end if XMLHttpRequest/XMLHTTP support

         return true; // indicate no errors
      }; // end method request


   this.handle = function()
      {
         // cycle through request objects to see
         // if any are ready with a response. we
         // keep looping even after we find one,
         // because it might not be the one that
         // fired the event (there could be
         // multiple that are ready.
         for (var i=0; i<this.aRequests.length; i++)
         {
            // if state is "complete"
            if (this.aRequests[i] != null
               && this.aRequests[i].readyState ==
                  this.READYSTATE_COMPLETED)
            {
               if (this.aRequests[i].status ==
                  this.HTTP_STATUS_OK)
               {

// TODO: remove
//document.getElementById('xmloutput').value = Sarissa.serialize(this.aRequests[i].responseXML);
document.getElementById('xmloutput').value = this.aRequests[i].responseText;

                  // pass this off to the xml parser
                  this.parseResponse(
                     this.aRequests[i].responseXML);

                  // remove object. this is
                  // important because Opera
                  // sometimes refires the
                  // readyStateChange event.
                  // plus, this might not be the
                  // one that fired the event, and
                  // this method might be running
                  // twice at the same time (setting
                  // this to null is about as ‘thread
                  // safe’ as we can get).
                  this.aRequests[i] = null;
               }
               else {
                   this.errors[this.errors.length] = 'Request failed with HTTP'
                      + 'response status code: '+ this.aRequests[i].status;
               }// end of HTTP OK
            } // end if completed
         } // end for
      }; // end method handle


   this.setRespHandler =
      function(sElementName, funcHandler)
      {
         // flip flag if wildcard
         if (sElementName == '*')
            this.isWildcardSet = true;
         // end if

         // add the element handler to the hashtable
         return this.hRespHandlers.put(
            sElementName, funcHandler);

         return true;
      }; // end method setRespHandler


   this.parseResponse = function(oNode)
      {
         if (!oNode) return;
         // base case (oNode is a leaf element)
         if (!oNode.hasChildNodes()) return;

         // else... recurse through children
         var children = oNode.childNodes;

         for (var i=0; i<children.length; i++)
         {
             // all nodes (element, attribute, text,
             // etc.) are returned as children, but
             // we only want to act on children that
             // are element nodes
             if (children[i].nodeType ==
                this.NODE_TYPE_ELEMENT)
             {
                // check to see if a handler exists
                // specifically for this element
                var elementName = children[i].nodeName;
                if ( this.hRespHandlers.containsKey(
                   elementName) )
                {
                   // if so, fire handler, and pass
                   // subtree starting with the node
                   // of interest

                   // retreive the handler
                   var funcHandler =
                     this.hRespHandlers.get(elementName);
                   // fire the handler and pass the
                   // subtree as its argument
                   funcHandler(children[i]);

                   // a match was found. conditionally
                   // recurse on the subtree.
                   // conditionally, because this subtree
                   // has already been handed off to one
                   // function.
                   if (this.recurseOnChildren)
                      this.parseResponse(children[i]);
                }
                else if (this.isWildcardSet)
                {
                   // retreive the handler
                   var funcHandler =
                      this.hRespHandlers.get('*');
                   // fire the handler and pass the
                   // subtree as its argument
                   funcHandler(children[i]);

                   // a match was found. conditionally
                   // recurse on the subtree.
                   // conditionally, because this subtree
                   // has already been handed off to one
                   // function.
                   if (this.recurseOnChildren)
                      this.parseResponse(children[i]);
                }
                else
                {
                   // not match yet found on this
                   // subtree. keep digging.
                   this.parseResponse(children[i]);
                }

             } // end if element node type
         } // end for loop of children
      }; // end method parseResponse



} // end class RemoteConnection

/**
 * This is a quick class to get around JavaScript's
 * poor support for associative arrays. Specifically,
 * without this class, properties of the object would
 * be included as keys in the associative array. This
 * class ensures that we deal only with array elements
 * we added to the array and not any properties of the
 * container we're using. See Side Bar 1.
 *
 * $Horde: incubator/Horde_Form/js/ajax.js,v 1.4 2006/03/02 23:50:19 chuck Exp $
 *
 * Copyright 2005 Matt Warden <mwarden@gmail.com>
 *
 * See the enclosed file COPYING for license information (LGPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/lgpl.html.
 *
 * @author Matt Warden <mwarden@gmail.com>
 * @access public
 */
function SimpleHashtable()
{
   this._data = new Object();
   this._keys = new Array();
} // end constructor


function SimpleHashtable_count()
{
    return this._keys.length;
}


/**
 * Lisp-style method to call callback on every
 * key-value pair of the array.
 *
 * @since 1.1
 * @author Marcus <m-lists@bristav.se>
 */
function SimpleHashtable_each(callback)
{
  for(var i=0 ; i<this._keys.length; i++) {
    var key = this._keys[i];
    callback(key, this.get(key));
  }
} // end function SimpleHashtable_each


function SimpleHashtable_getKey(raw)
{
   // python style naming convention,
   // to avoid conflicts with actual
   // attributes (see Side Bar 1).
   return '__'+ raw +'__';
} // end function SimpleHashtable_getKey


function SimpleHashtable_get(nam)
{
   var key = this.getKey(nam);
   // retreive value if exists, else null
   var val = (this._data[key]) ? this._data[key] : null;
   return val;
} // end function SimpleHashtable_get


function SimpleHashtable_put(nam, val)
{
   // if missing arg
   if (!nam) return false;

   var key = this.getKey(nam);

   // if key doesn't already exist add
   // to keys array.
   var exists = true;
   if (!this._data[key])
   {
      exists = false;
      this._keys[this._keys.length] = nam;
   }

   // return old value if set, or else null
   var oldval = exists ? this._data[key] : null;

   this._data[key] = val;

   return oldval;
} // end function SimpleHashtable_put


function SimpleHashtable_keys()
{
   // return a copy of the array, otherwise
   // the user of this class could accidentally
   // modify our array, since we'd be passing
   // them a reference.
   return this._keys.slice(0);
} // end function SimpleHashtable_keys


function SimpleHashtable_containsKey(nam)
{
   // get() returns null if not found
   // or if found and is null. In both
   // conditions, we want to return false.
   // Otherwise, we want to return true.
   return (this.get(nam) != null);
} // end function SimpleHashtable_containsKey

// creating the class this way is more memory efficient
// if we are using multiple instances of this data
// structure.
SimpleHashtable.prototype.getKey = SimpleHashtable_getKey;
SimpleHashtable.prototype.get = SimpleHashtable_get;
SimpleHashtable.prototype.put = SimpleHashtable_put;
SimpleHashtable.prototype.keys = SimpleHashtable_keys;
SimpleHashtable.prototype.count = SimpleHashtable_count;
SimpleHashtable.prototype.containsKey = SimpleHashtable_containsKey;
SimpleHashtable.prototype.each = SimpleHashtable_each;
