/**
 * Static functionality for incremental validation of Horde_Form's.
 *
 * $Horde: incubator/Horde_Form/js/incrementalvalidator.js,v 1.6 2006/03/02 23:50:19 chuck Exp $
 *
 * Copyright 2005 Matt Warden <mwarden@gmail.com>
 *
 * See the enclosed file COPYING for license information (LGPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/lgpl.html.
 *
 * @author Matt Warden <mwarden@gmail.com>
 * @access public
 */

// TODO: is there any point in making this more object-oriented? I didn't
// because it's all event-driven.


/**
 * The remote connection (for AJAX-style communication).
 *
 * @var RemoteConnection
 */
var __hfiv_conn = new RemoteConnection(false, true);
__hfiv_conn.setRespHandler('validation-result', __hfiv_handleValResult);


var __hfiv_validationurl = window.location.href;
__hfiv_validationurl += (__hfiv_validationurl.indexOf('?') >= 0) ? '&' : '?';
__hfiv_validationurl += '__hfiv_incremental=true';

var __hfiv_postxmldoc = null;

/**
 * Storage of IDs of all elements that have been touched by the user.
 *
 * @var arry
 */
var __hfiv_touched = [];


/**
 * Initiates incremental validation of the form and sets the next time
 * to validate.
 *
 * @param DOMDocument XML DOM document representing last state of the form.
 *      if null, the value stored in __hfiv_postxmldoc is used.
 * @param int number of milliseconds to wait until next validation
 * @return void
 */
function __hfiv_validate(xmldoc, revalidateDelay)
{
    if (xmldoc) __hfiv_postxmldoc = xmldoc;

    // warning: slope is *not* linear!
    //var delaySlope = 1.25;
    var delaySlope = 1.0;

    // only if touched
    if (__hfiv_postxmldoc != null && __hfiv_touched.length > 0) {
        var valnodes = __hfiv_postxmldoc.documentElement.getElementsByTagName('value');
        for (var i = 0; i < valnodes.length; i++) {
            if (valnodes[i].firstChild) {
                valnodes[i].removeChild(valnodes[i].firstChild);
            }
            // go ahead and set parent's touched to false to allow for untouching
            // if we ever need it
            valnodes[i].parentNode.setAttribute('touched', 'false');
        }

        var namenodes = __hfiv_postxmldoc.documentElement.getElementsByTagName('name');
        // update the values for touched elements
        // for each field marked as touched
        for (var i = 0; i < __hfiv_touched.length; ++i) {
            var touchedId = __hfiv_touched[i];
            // for each "name" element
            for (var j = 0; j < namenodes.length; ++j) {
                if (namenodes[j] && namenodes[j].firstChild
                    && namenodes[j].firstChild.nodeValue == touchedId) {
                    // namenodes[j] is the node that corresponds to this
                    // form element's <name> node, so its parent is the correct
                    // <variable> node
                    var varnode = namenodes[j].parentNode;
                    // get the document element that corresponds to this variable
                    // node
                    var frmelem = document.getElementById(touchedId);
                    // get the value(s) regardless of field type.
                    // returns an array of values or false if unable to get a value
                    var currentVal = __hfiv_getFieldValue(frmelem);
                    if (currentVal) {
                        // for each value of the field
                        for (var k=0; k<currentVal.length; k++) {
                            var valnode;

                            if (k > 0) {
                                // create another value node
                                valnode = __hfiv_postxmldoc.createElement('value');
                            }
                            else {
                                // there is already one empty value node in the document.
                                // it signifies an empty value (just like initializing
                                // a variable).
                                valnode = varnode.getElementsByTagName('value')[0];
                            }

                            var txtnode = __hfiv_postxmldoc.createTextNode(currentVal[k]);
                            valnode.appendChild(txtnode);

                            if (k > 0) {
                                // we need to append the valnode to the document, because it
                                // is not already in the tree (operations are faster if we
                                // append children before appending to DOM)
                                varnode.appendChild(valnode);
                            }
                        } // end value for loop
                    } // end if value found
                    // also make sure touched attr is true
                    varnode.setAttribute('touched', 'true');
                    break;
                }
            } // end inner for
        } // end outer for

        __hfiv_conn.request(__hfiv_validationurl, 'post', __hfiv_postxmldoc)
    } // end if touched


    setTimeout('__hfiv_validate(false, '+ revalidateDelay +')', revalidateDelay * delaySlope)
} // end function __hfiv_validate()


function __hfiv_handleValResult(node)
{
    // sorry, can't support you if you don't support DOM1
    if (!document.getElementsByTagName) return;

    var parent = node.parentNode;
    if (!parent) return; // que pasa? can't handle an orphaned validation-result

    var targetField = parent.getAttribute('for');
    if (!targetField) return; // que pasa? can't handle an orphaned validation-result

    // get reference to target field
    var target = document.getElementById(targetField);
    if (!target) {
        // sorry. (TODO: add support for .name ??? Might not be worth it.)
        return;
    }

    var currentVal = __hfiv_getFieldValue(target);

    // get the value that was validated. if the value has since changed,
    // these results are stale.
    var elemValue = node.getElementsByTagName('value');
    // there should only be one result
    if (elemValue) elemValue = elemValue.item(0);
    var validatedValue = false;
    if (elemValue && elemValue.firstChild && elemValue.firstChild.nodeValue) {
        validatedValue = elemValue.firstChild.nodeValue;
    }
    // TODO: add support for multi-value fields
    if (validatedValue && validatedValue != currentVal) {
        return;
    }

    // result of the validation
    var elemValid = node.getElementsByTagName('isvalid');
    // there should only be one result
    if (elemValid) elemValid = elemValid.item(0);
    var isvalid = false;
    if (elemValid && elemValid.firstChild && elemValid.firstChild.nodeValue) {
        isvalid = (elemValid.firstChild.nodeValue == 'true');
    }

    // is there an error message?
    var errormsg = '';
    if (!isvalid) {
        // check to see if there is an error message
        var elemErrMsg = node.getElementsByTagName('errormsg');
        // there should only be one result
        if (elemErrMsg) elemErrMsg = elemErrMsg.item(0);
        if (elemErrMsg && elemErrMsg.firstChild && elemErrMsg.firstChild.nodeValue) {
            errormsg = elemErrMsg.firstChild.nodeValue;
        }
    } // end if invalid

    // manipulate the DOM as generally as possible

    // find parent DIV
    var targetParentDIV = __hfiv_getParentDiv(target);
    if (!targetParentDIV) {
        // if we didn't find the parent div, there's not much we can do as far
        // as DOM manipulations go
        return;
    }

    if (!isvalid) {
        cssjs('add',targetParentDIV,'form-error');

        // remove any message nodes from before if msg different
        var bMsgExists = false;
        var bMsgIsDifferent = false;
        var children = targetParentDIV.childNodes;
        for (var i=0; i<children.length; i++) {
            if (children[i].nodeName
                && children[i].nodeName.toUpperCase() == 'P'
                && cssjs('check', children[i], 'form-error')) {
                bMsgExists = true;
                var grandchildren = children[i].childNodes;
                for (var j=0; j<grandchildren.length; j++) {
                    // get the text node child
                    if (grandchildren[j].nodeType == 3) {
                        var currentMsg = grandchildren[j].nodeValue;
                        if (currentMsg != errormsg) {
                            // error msg is different. remove p node.
                            bMsgIsDifferent = true;
                            targetParentDIV.removeChild(children[i]);
                            break;
                        }
                    } // end if
                } // end for
            }
        } // end for

        if ((!bMsgExists || bMsgIsDifferent) && document.createElement && errormsg!='') {
            var errorp = document.createElement('p');
            errorp.appendChild(document.createTextNode(errormsg));
            cssjs('add', errorp, 'form-error');

            // find first child that is a tag
            var fchild = null; // first tag child
            var children = targetParentDIV.childNodes;
            for (var i=0; i<children.length; i++) {
                // if this is the first tag we've seen
                if (fchild == null && children[i].nodeType == 1) {
                    fchild = children[i];
                    break;
                }
            }

            // no replacing
            if (fchild == null) {
                // should never happen, but...
                targetParentDIV.appendChild(errorp);
            }
            else {
                targetParentDIV.insertBefore(errorp, fchild);
            }
        } // end if error message
    }
    else {
        // remove form error classification
        cssjs('remove',targetParentDIV,'form-error');
        // remove any message nodes from before
        var children = targetParentDIV.childNodes;
        for (var i=0; i<children.length; i++) {
            if (children[i].nodeName
                && children[i].nodeName.toUpperCase() == 'P'
                && cssjs('check', children[i], 'form-error')) {
                targetParentDIV.removeChild(children[i]);
            }
        } // end for
    } // end if invalid
} // end function __hfiv_handleVariable


/**
 * Returns an array of values for the form field, regardless of type.
 * If the type is such that the field has only one value, an array is
 * still returned (it will have a single element).
 *
 * @param Object field - the form field whose values should be returned
 * @return Array|false an array of values for the form field referenced
 *      by field, or false if no value could be determined.
 */
function __hfiv_getFieldValue(field)
{
    var retval = false;

    if (document.getElementsByTagName && field) {
        try {
            // .type has been supported since JS1.1
            // if .type isn't available, I'm not bothering
            // any further, because the rest of this script
            // isn't going to work anyway (DOM1 needs to be
            // supported). But, if this gets pulled out into
            // some generalized function, we might want to
            // include some support for JS < 1.1 (or at least
            // know it won't work). hence this note.
            if (field.type) {
                switch (field.type.toLowerCase()) {
                    case 'button': case 'reset': case 'submit':
                        retval = new Array(field.value);
                        break;
                    case 'checkbox': case 'radio':
                        var groupname = field.name;
                        var parentDiv = __hfiv_getParentDiv(field);
                        if (!parentDiv) return;
                        var siblings = parentDiv.getElementsByTagName('input');
                        for (var i=0; i<siblings.length; i++) {
                            if (siblings[i].name && groupname == siblings[i].name) {
                                if (sublings[i].checked) {
                                    if (retval) {
                                        retval[retval.length] = siblings[i].value;
                                    }
                                    else {
                                        retval = new Array(siblings[i].value);
                                    }
                                } // end if checked
                            } // end if in group
                        } // end for
                        break;
                    case 'hidden':
                        retval = new Array(field.value);
                        break;
                    case 'file':
                        retval = new Array(field.value);
                        break;
                    case 'password':
                        // check if password-confirm

                        // should this just check for a sibling input field?
                        // one might argue that I am using 'class' like 'type' in this
                        // instance. Then again, 'class' is a grouping, and having the
                        // class of form-input-passwordconfirm groups the two inputs
                        // as desired.

                        // if field or parentDiv have form-input-passwordconfirm classname
                        // (either, because we plan on moving the class names to the DIV
                        // -- once we do this, we could change the order of the conditions
                        // to optimize -- or we could simply remove the first condition)
                        var parentDiv = __hfiv_getParentDiv(field);
                        if (cssjs('check', field, 'form-input-passwordconfirm')
                            || (parentDiv && cssjs('check', parentDiv, 'form-input-passwordconfirm'))) {
                            // this is a password-confirm control. in Horde_Form terms,
                            // this is a single variable.

                            // get the id of the "confirm" field.
                            var confirmId = __hfiv_getConfirmId(field.id);
                            var confirmField = document.getElementById(confirmId)
                            if (confirmField) {
                                retval = new Array(field.value, confirmField.value);
                            }
                            else {
                                retval = new Array(field.value);
                            }
                        }
                        else {
                            retval = new Array(field.value);
                        } // end if
                        break;
                    case 'select': case 'select-one':
                        var idx = field.selectedIndex;
                        retval = new Array(field.options[idx].value);
                        break;
                    case 'select-multiple':
                        var opts = field.options;
                        for (var i=0; i<opts.length; i++) {
                            if (opts.selected) {
                                if (retval) {
                                    retval[retval.length] = siblings[i].value;
                                }
                                else {
                                    retval = new Array(siblings[i].value);
                                }
                            }
                        }
                        break;
                    case 'text':
                        // check if email-confirm

                        // should this just check for a sibling input field?
                        // (see password-confirm note)

                        // if field or parentDiv have form-input-emailconfirm classname
                        // (either, because we plan on moving the class names to the DIV
                        // -- once we do this, we could change the order of the conditions
                        // to optimize -- or we could simply remove the first condition)
                        var parentDiv = __hfiv_getParentDiv(field);
                        if (cssjs('check', field, 'form-input-emailconfirm')
                            || (parentDiv && cssjs('check', parentDiv, 'form-input-emailconfirm'))) {
                            // this is a email-confirm control. in Horde_Form terms,
                            // this is a single variable.

                            // get the id of the "confirm" field.
                            var confirmId = __hfiv_getConfirmId(field.id);
                            var confirmField = document.getElementById(confirmId)
                            if (confirmField) {
                                retval = new Array(field.value, confirmField.value);
                            }
                            else {
                                retval = new Array(field.value);
                            }
                        }
                        else {
                            retval = new Array(field.value);
                        } // end if
                        break;
                    case 'textarea':
                        retval = new Array(field.value);
                        break;
                } // end switch
            } // end if field.type
        }
        catch (e) {} // end try
    } // end if field

    return retval;
}


function __hfiv_getParentDiv(field)
{
    if (!field) return false;
    var parentDiv;
    while ((parentDiv = field.parentNode) != null) {
        if (parentDiv.nodeName
        && parentDiv.nodeName.toUpperCase() == 'DIV') {
            break;
        }
    }
    return parentDiv;
}


function __hfiv_getConfirmId(fieldName)
{
    return (fieldName +'-confirm-');
}



/**
 * Marks the element as touched. This is attached as an event handler.
 *
 */
function __hfiv_touch(e)
{
    // if e parameter isn't defined, use IE's 'event' var
    e = e || event;
    var id = false;
    if (e.srcElement || e.target) {
        id = (e.srcElement || e.target).id;
    }
    if (!id) id = this.id;
    if (!id) id = this.name;
    if (!id) return;

    __hfiv_touched.push(id);
}
